@@ -31,6 +31,7 @@
 #include "gtlsinputstream-gnutls.h"
 #include "gtlsoutputstream-gnutls.h"
 #include "gtlsserverconnection-gnutls.h"
+#include "gnutls-marshal.h"
 #include <glib/gi18n-lib.h>
 
 static void g_tls_connection_gnutls_get_property (GObject    *object,
@@ -82,9 +83,13 @@ static gboolean g_tls_connection_gnutls_initable_init       (GInitable       *in
 							     GCancellable    *cancellable,
 							     GError         **error);
 
+static void g_tls_connection_gnutls_init_priorities (void);
+
 G_DEFINE_ABSTRACT_TYPE_WITH_CODE (GTlsConnectionGnutls, g_tls_connection_gnutls, G_TYPE_TLS_CONNECTION,
 				  G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE,
-							 g_tls_connection_gnutls_initable_iface_init));
+							 g_tls_connection_gnutls_initable_iface_init);
+				  g_tls_connection_gnutls_init_priorities ();
+				  );
 
 
 enum
@@ -92,7 +97,13 @@ enum
   PROP_0,
   PROP_BASE_IO_STREAM,
   PROP_REQUIRE_CLOSE_NOTIFY,
-  PROP_REHANDSHAKE_MODE
+  PROP_REHANDSHAKE_MODE,
+  PROP_USE_SYSTEM_CERTDB,
+  PROP_DATABASE,
+  PROP_CERTIFICATE,
+  PROP_INTERACTION,
+  PROP_PEER_CERTIFICATE,
+  PROP_PEER_CERTIFICATE_ERRORS
 };
 
 struct _GTlsConnectionGnutlsPrivate
@@ -101,20 +112,30 @@ struct _GTlsConnectionGnutlsPrivate
   GPollableInputStream *base_istream;
   GPollableOutputStream *base_ostream;
 
-  GList *ca_list;
   gnutls_certificate_credentials creds;
   gnutls_session session;
+
+  GTlsCertificate *certificate, *peer_certificate;
+  GTlsCertificateFlags peer_certificate_errors;
   gboolean require_close_notify;
   GTlsRehandshakeMode rehandshake_mode;
+  gboolean is_system_certdb;
+  GTlsDatabase *database;
+  gboolean database_is_unset;
   gboolean need_handshake, handshaking, ever_handshaked;
   gboolean closing;
 
   GInputStream *tls_istream;
   GOutputStream *tls_ostream;
 
+  GTlsInteraction *interaction;
+
   GError *error;
   GCancellable *cancellable;
-  gboolean blocking, eof;
+  gboolean blocking;
+#ifndef GNUTLS_E_PREMATURE_TERMINATION
+  gboolean eof;
+#endif
   GIOCondition internal_direction;
 };
 
@@ -144,6 +165,12 @@ g_tls_connection_gnutls_class_init (GTlsConnectionGnutlsClass *klass)
   g_object_class_override_property (gobject_class, PROP_BASE_IO_STREAM, "base-io-stream");
   g_object_class_override_property (gobject_class, PROP_REQUIRE_CLOSE_NOTIFY, "require-close-notify");
   g_object_class_override_property (gobject_class, PROP_REHANDSHAKE_MODE, "rehandshake-mode");
+  g_object_class_override_property (gobject_class, PROP_USE_SYSTEM_CERTDB, "use-system-certdb");
+  g_object_class_override_property (gobject_class, PROP_DATABASE, "database");
+  g_object_class_override_property (gobject_class, PROP_CERTIFICATE, "certificate");
+  g_object_class_override_property (gobject_class, PROP_INTERACTION, "interaction");
+  g_object_class_override_property (gobject_class, PROP_PEER_CERTIFICATE, "peer-certificate");
+  g_object_class_override_property (gobject_class, PROP_PEER_CERTIFICATE_ERRORS, "peer-certificate-errors");
 }
 
 static void
@@ -162,6 +189,44 @@ g_tls_connection_gnutls_init (GTlsConnectionGnutls *gnutls)
 				       GNUTLS_VERIFY_ALLOW_X509_V1_CA_CRT);
 
   gnutls->priv->need_handshake = TRUE;
+
+  gnutls->priv->database_is_unset = TRUE;
+  gnutls->priv->is_system_certdb = TRUE;
+}
+
+static gnutls_priority_t priorities[2][2];
+
+static void
+g_tls_connection_gnutls_init_priorities (void)
+{
+  /* First field is "ssl3 only", second is "allow unsafe rehandshaking" */
+
+  gnutls_priority_init (&priorities[FALSE][FALSE],
+			"NORMAL:%COMPAT",
+			NULL);
+  gnutls_priority_init (&priorities[TRUE][FALSE],
+			"NORMAL:%COMPAT:!VERS-TLS1.2:!VERS-TLS1.1:!VERS-TLS1.0",
+			NULL);
+  gnutls_priority_init (&priorities[FALSE][TRUE],
+			"NORMAL:%COMPAT:%UNSAFE_RENEGOTIATION",
+			NULL);
+  gnutls_priority_init (&priorities[TRUE][TRUE],
+			"NORMAL:%COMPAT:!VERS-TLS1.2:!VERS-TLS1.1:!VERS-TLS1.0:%UNSAFE_RENEGOTIATION",
+			NULL);
+}
+
+static void
+g_tls_connection_gnutls_set_handshake_priority (GTlsConnectionGnutls *gnutls)
+{
+  gboolean use_ssl3, unsafe_rehandshake;
+
+  if (G_IS_TLS_CLIENT_CONNECTION (gnutls))
+    use_ssl3 = g_tls_client_connection_get_use_ssl3 (G_TLS_CLIENT_CONNECTION (gnutls));
+  else
+    use_ssl3 = FALSE;
+  unsafe_rehandshake = (gnutls->priv->rehandshake_mode == G_TLS_REHANDSHAKE_UNSAFELY);
+  gnutls_priority_set (gnutls->priv->session,
+		       priorities[use_ssl3][unsafe_rehandshake]);
 }
 
 static gboolean
@@ -170,9 +235,6 @@ g_tls_connection_gnutls_initable_init (GInitable     *initable,
 				       GError       **error)
 {
   GTlsConnectionGnutls *gnutls = G_TLS_CONNECTION_GNUTLS (initable);
-  gboolean client, use_ssl3 = FALSE;
-  gnutls_x509_crt_t *cas;
-  int num_cas;
   int status;
 
   g_return_val_if_fail (gnutls->priv->base_istream != NULL &&
@@ -183,31 +245,6 @@ g_tls_connection_gnutls_initable_init (GInitable     *initable,
    */
   g_tls_connection_gnutls_get_session (gnutls);
 
-  client = G_IS_TLS_CLIENT_CONNECTION (gnutls);
-  if (client)
-      g_object_get (G_OBJECT (gnutls), "use-ssl3", &use_ssl3, NULL);
-  if (use_ssl3)
-    {
-      status = gnutls_priority_set_direct (gnutls->priv->session,
-					   "NORMAL:!VERS-TLS1.2:!VERS-TLS1.1:!VERS-TLS1.0",
-					   NULL);
-    }
-  else
-    {
-      status = gnutls_priority_set_direct (gnutls->priv->session,
-					   "NORMAL", NULL);
-    }
-  if (status != 0)
-    {
-      g_set_error (error, G_TLS_ERROR, G_TLS_ERROR_MISC,
-		   _("Could not create TLS connection: %s"),
-		   gnutls_strerror (status));
-      return FALSE;
-    }
-
-  g_tls_backend_gnutls_get_system_ca_list_gnutls (&cas, &num_cas);
-  gnutls_certificate_set_x509_trust (gnutls->priv->creds, cas, num_cas);
-
   status = gnutls_credentials_set (gnutls->priv->session,
 				   GNUTLS_CRD_CERTIFICATE,
 				   gnutls->priv->creds);
@@ -219,6 +256,11 @@ g_tls_connection_gnutls_initable_init (GInitable     *initable,
       return FALSE;
     }
 
+  /* Some servers (especially on embedded devices) use tiny keys that
+   * gnutls will reject by default. We want it to accept them.
+   */
+  gnutls_dh_set_prime_bits (gnutls->priv->session, 256);
+
   gnutls_transport_set_push_function (gnutls->priv->session,
 				      g_tls_connection_gnutls_push_func);
   gnutls_transport_set_pull_function (gnutls->priv->session,
@@ -250,6 +292,15 @@ g_tls_connection_gnutls_finalize (GObject *object)
   if (connection->priv->creds)
     gnutls_certificate_free_credentials (connection->priv->creds);
 
+  if (connection->priv->database)
+    g_object_unref (connection->priv->database);
+  if (connection->priv->certificate)
+    g_object_unref (connection->priv->certificate);
+  if (connection->priv->peer_certificate)
+    g_object_unref (connection->priv->peer_certificate);
+
+  g_clear_object (&connection->priv->interaction);
+
   if (connection->priv->error)
     g_error_free (connection->priv->error);
 
@@ -263,6 +314,7 @@ g_tls_connection_gnutls_get_property (GObject    *object,
 				      GParamSpec *pspec)
 {
   GTlsConnectionGnutls *gnutls = G_TLS_CONNECTION_GNUTLS (object);
+  GTlsBackend *backend;
 
   switch (prop_id)
     {
@@ -278,6 +330,36 @@ g_tls_connection_gnutls_get_property (GObject    *object,
       g_value_set_enum (value, gnutls->priv->rehandshake_mode);
       break;
 
+    case PROP_USE_SYSTEM_CERTDB:
+      g_value_set_boolean (value, gnutls->priv->is_system_certdb);
+      break;
+
+    case PROP_DATABASE:
+      if (gnutls->priv->database_is_unset)
+        {
+          backend = g_tls_backend_get_default ();
+          gnutls->priv->database =  g_tls_backend_get_default_database (backend);
+          gnutls->priv->database_is_unset = FALSE;
+        }
+      g_value_set_object (value, gnutls->priv->database);
+      break;
+
+    case PROP_CERTIFICATE:
+      g_value_set_object (value, gnutls->priv->certificate);
+      break;
+
+    case PROP_INTERACTION:
+      g_value_set_object (value, gnutls->priv->interaction);
+      break;
+
+    case PROP_PEER_CERTIFICATE:
+      g_value_set_object (value, gnutls->priv->peer_certificate);
+      break;
+
+    case PROP_PEER_CERTIFICATE_ERRORS:
+      g_value_set_flags (value, gnutls->priv->peer_certificate_errors);
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
     }
@@ -290,9 +372,10 @@ g_tls_connection_gnutls_set_property (GObject      *object,
 				      GParamSpec   *pspec)
 {
   GTlsConnectionGnutls *gnutls = G_TLS_CONNECTION_GNUTLS (object);
-  GList *list, *copy, *c;
   GInputStream *istream;
   GOutputStream *ostream;
+  gboolean system_certdb;
+  GTlsBackend *backend;
 
   switch (prop_id)
     {
@@ -326,6 +409,38 @@ g_tls_connection_gnutls_set_property (GObject      *object,
       gnutls->priv->rehandshake_mode = g_value_get_enum (value);
       break;
 
+    case PROP_USE_SYSTEM_CERTDB:
+      system_certdb = g_value_get_boolean (value);
+      if (system_certdb != gnutls->priv->is_system_certdb)
+        {
+          g_clear_object (&gnutls->priv->database);
+          if (system_certdb)
+            {
+              backend = g_tls_backend_get_default ();
+              gnutls->priv->database = g_tls_backend_get_default_database (backend);
+            }
+          gnutls->priv->is_system_certdb = system_certdb;
+        }
+      break;
+
+    case PROP_DATABASE:
+      g_clear_object (&gnutls->priv->database);
+      gnutls->priv->database = g_value_dup_object (value);
+      gnutls->priv->is_system_certdb = FALSE;
+      gnutls->priv->database_is_unset = FALSE;
+      break;
+
+    case PROP_CERTIFICATE:
+      if (gnutls->priv->certificate)
+	g_object_unref (gnutls->priv->certificate);
+      gnutls->priv->certificate = g_value_dup_object (value);
+      break;
+
+    case PROP_INTERACTION:
+      g_clear_object (&gnutls->priv->interaction);
+      gnutls->priv->interaction = g_value_dup_object (value);
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
     }
@@ -362,10 +477,6 @@ g_tls_connection_gnutls_get_certificate (GTlsConnectionGnutls *gnutls,
   GTlsCertificate *cert;
 
   cert = g_tls_connection_get_certificate (G_TLS_CONNECTION (gnutls));
-  if (cert)
-    g_object_ref (cert);
-  else
-    cert = g_tls_connection_emit_need_certificate (G_TLS_CONNECTION (gnutls));
 
   st->type = GNUTLS_CRT_X509;
   if (cert)
@@ -377,58 +488,12 @@ g_tls_connection_gnutls_get_certificate (GTlsConnectionGnutls *gnutls,
       st->cert.x509[0] = g_tls_certificate_gnutls_copy_cert (gnutlscert);
       st->key.x509 = g_tls_certificate_gnutls_copy_key (gnutlscert);
       st->deinit_all = TRUE;
-
-      g_object_unref (cert);
     }
   else
     st->ncerts = 0;
 }
 
-static const struct {
-  int gnutls_flag;
-  GTlsCertificateFlags gtls_flag;
-} flags_map[] = {
-  { GNUTLS_CERT_SIGNER_NOT_FOUND | GNUTLS_CERT_SIGNER_NOT_CA, G_TLS_CERTIFICATE_UNKNOWN_CA },
-  { GNUTLS_CERT_NOT_ACTIVATED, G_TLS_CERTIFICATE_NOT_ACTIVATED },
-  { GNUTLS_CERT_EXPIRED, G_TLS_CERTIFICATE_EXPIRED },
-  { GNUTLS_CERT_REVOKED, G_TLS_CERTIFICATE_REVOKED },
-  { GNUTLS_CERT_INSECURE_ALGORITHM, G_TLS_CERTIFICATE_INSECURE }
-};
-static const int flags_map_size = G_N_ELEMENTS (flags_map);
-
-GTlsCertificateFlags
-g_tls_connection_gnutls_validate_peer (GTlsConnectionGnutls *gnutls)
-{
-  int status, i;
-  GTlsCertificateFlags gtls_errors;
-
-  status = gnutls_certificate_verify_peers (gnutls->priv->session);
-
-  /* Convert GNUTLS status to GTlsCertificateFlags. GNUTLS sets
-   * GNUTLS_CERT_INVALID if it sets any other flag, so we want to
-   * strip that out unless it's the only flag set. Then we convert
-   * specific flags we recognize, and if there are any flags left over
-   * at the end, we add G_TLS_CERTIFICATE_GENERIC_ERROR.
-   */
-  gtls_errors = 0;
-
-  if (status != GNUTLS_CERT_INVALID)
-    status = status & ~GNUTLS_CERT_INVALID;
-  for (i = 0; i < flags_map_size && status != 0; i++)
-    {
-      if (status & flags_map[i].gnutls_flag)
-	{
-	  status &= ~flags_map[i].gnutls_flag;
-	  gtls_errors |= flags_map[i].gtls_flag;
-	}
-    }
-  if (status)
-    gtls_errors |= G_TLS_CERTIFICATE_GENERIC_ERROR;
-
-  return gtls_errors;
-}
-
-void
+static void
 begin_gnutls_io (GTlsConnectionGnutls  *gnutls,
 		 gboolean               blocking,
 		 GCancellable          *cancellable)
@@ -439,10 +504,9 @@ begin_gnutls_io (GTlsConnectionGnutls  *gnutls,
   g_clear_error (&gnutls->priv->error);
 }
 
-int
+static int
 end_gnutls_io (GTlsConnectionGnutls  *gnutls,
 	       int                    status,
-	       const char            *generic_error,
 	       GError               **error)
 {
   gnutls->priv->cancellable = NULL;
@@ -477,26 +541,34 @@ end_gnutls_io (GTlsConnectionGnutls  *gnutls,
     }
   else if (status == GNUTLS_E_REHANDSHAKE)
     {
+      if (gnutls->priv->rehandshake_mode == G_TLS_REHANDSHAKE_NEVER)
+	{
+	  g_set_error_literal (error, G_TLS_ERROR, G_TLS_ERROR_MISC,
+			       _("Peer requested illegal TLS rehandshake"));
+	  return GNUTLS_E_PULL_ERROR;
+	}
+
       gnutls->priv->need_handshake = TRUE;
       return status;
     }
-  else if (status == GNUTLS_E_UNEXPECTED_PACKET_LENGTH)
+  else if (
+#ifdef GNUTLS_E_PREMATURE_TERMINATION
+	   status == GNUTLS_E_PREMATURE_TERMINATION
+#else
+	   status == GNUTLS_E_UNEXPECTED_PACKET_LENGTH && gnutls->priv->eof
+#endif
+	   )
     {
-      if (gnutls->priv->eof)
+      if (gnutls->priv->require_close_notify)
 	{
-	  if (gnutls->priv->require_close_notify)
-	    {
-	      g_set_error_literal (error, G_TLS_ERROR, G_TLS_ERROR_EOF,
-				   _("TLS connection closed unexpectedly"));
-	      return status;
-	    }
-	  else
-	    return 0;
+	  g_set_error_literal (error, G_TLS_ERROR, G_TLS_ERROR_EOF,
+			       _("TLS connection closed unexpectedly"));
+	  return status;
 	}
+      else
+	return 0;
     }
 
-  g_set_error (error, G_TLS_ERROR, G_TLS_ERROR_MISC,
-	       generic_error, gnutls_strerror (status));
   return status;
 }
 
@@ -508,12 +580,21 @@ end_gnutls_io (GTlsConnectionGnutls  *gnutls,
   } while ((ret == GNUTLS_E_AGAIN ||			\
             ret == GNUTLS_E_WARNING_ALERT_RECEIVED) &&	\
            !gnutls->priv->error);			\
-  ret = end_gnutls_io (gnutls, ret, errmsg, error)
+  ret = end_gnutls_io (gnutls, ret, error);		\
+  if (ret < 0 && ret != GNUTLS_E_REHANDSHAKE && error && !*error) \
+    {							\
+      g_set_error (error, G_TLS_ERROR, G_TLS_ERROR_MISC,\
+                   errmsg, gnutls_strerror (ret));	\
+    }							\
+  ;
 
 gboolean
 g_tls_connection_gnutls_check (GTlsConnectionGnutls  *gnutls,
 			       GIOCondition           condition)
 {
+  if (!gnutls->priv->internal_direction)
+    return TRUE;
+
   if (gnutls->priv->handshaking || gnutls->priv->closing)
     condition = gnutls->priv->internal_direction;
 
@@ -530,7 +611,6 @@ typedef struct {
   GObject              *stream;
 
   GSource              *child_source;
-  GIOCondition          base_direction;
   GIOCondition          current_direction;
 } GTlsConnectionGnutlsSource;
 
@@ -548,15 +628,24 @@ gnutls_source_check (GSource *source)
   return FALSE;
 }
 
-static void
-gnutls_source_sync_child_source (GTlsConnectionGnutlsSource *gnutls_source,
-				 GIOCondition                direction)
+static gboolean
+gnutls_source_sync_child_source (GTlsConnectionGnutlsSource *gnutls_source)
 {
   GTlsConnectionGnutls *gnutls = gnutls_source->gnutls;
   GSource *source = (GSource *)gnutls_source;
+  GIOCondition direction;
+
+  if (gnutls->priv->handshaking || gnutls->priv->closing)
+    direction = gnutls->priv->internal_direction;
+  else if (!gnutls_source->stream)
+    return FALSE;
+  else if (G_IS_TLS_INPUT_STREAM_GNUTLS (gnutls_source->stream))
+    direction = G_IO_IN;
+  else
+    direction = G_IO_OUT;
 
   if (direction == gnutls_source->current_direction)
-    return;
+    return TRUE;
 
   if (gnutls_source->child_source)
     {
@@ -572,6 +661,7 @@ gnutls_source_sync_child_source (GTlsConnectionGnutlsSource *gnutls_source,
   g_source_set_dummy_callback (gnutls_source->child_source);
   g_source_add_child_source (source, gnutls_source->child_source);
   gnutls_source->current_direction = direction;
+  return TRUE;
 }
 
 static gboolean
@@ -581,16 +671,11 @@ gnutls_source_dispatch (GSource     *source,
 {
   GPollableSourceFunc func = (GPollableSourceFunc)callback;
   GTlsConnectionGnutlsSource *gnutls_source = (GTlsConnectionGnutlsSource *)source;
-  GTlsConnectionGnutls *gnutls = gnutls_source->gnutls;
   gboolean ret;
 
   ret = (*func) (gnutls_source->stream, user_data);
   if (ret)
-    {
-      GIOCondition direction = gnutls->priv->internal_direction ? gnutls->priv->internal_direction : gnutls_source->base_direction;
-
-      gnutls_source_sync_child_source (gnutls_source, direction);
-    }
+    ret = gnutls_source_sync_child_source (gnutls_source);
 
   return ret;
 }
@@ -606,12 +691,38 @@ gnutls_source_finalize (GSource *source)
     g_source_unref (gnutls_source->child_source);
 }
 
+static gboolean
+g_tls_connection_gnutls_source_closure_callback (GObject  *stream,
+						 gpointer  data)
+{
+  GClosure *closure = data;
+
+  GValue param = { 0, };
+  GValue result_value = { 0, };
+  gboolean result;
+
+  g_value_init (&result_value, G_TYPE_BOOLEAN);
+
+  g_value_init (&param, G_TYPE_OBJECT);
+  g_value_set_object (&param, stream);
+
+  g_closure_invoke (closure, &result_value, 1, &param, NULL);
+
+  result = g_value_get_boolean (&result_value);
+  g_value_unset (&result_value);
+  g_value_unset (&param);
+
+  return result;
+}
+
 static GSourceFuncs gnutls_source_funcs =
 {
   gnutls_source_prepare,
   gnutls_source_check,
   gnutls_source_dispatch,
-  gnutls_source_finalize
+  gnutls_source_finalize,
+  (GSourceFunc)g_tls_connection_gnutls_source_closure_callback,
+  (GSourceDummyMarshal)_gnutls_marshal_BOOLEAN__VOID,
 };
 
 GSource *
@@ -626,17 +737,11 @@ g_tls_connection_gnutls_create_source (GTlsConnectionGnutls  *gnutls,
   g_source_set_name (source, "GTlsConnectionGnutlsSource");
   gnutls_source = (GTlsConnectionGnutlsSource *)source;
   gnutls_source->gnutls = g_object_ref (gnutls);
-  gnutls_source->base_direction = condition & (G_IO_IN | G_IO_OUT);
-  if (gnutls_source->base_direction == G_IO_IN)
+  if (condition & G_IO_IN)
     gnutls_source->stream = G_OBJECT (gnutls->priv->tls_istream);
-  else if (gnutls_source->base_direction == G_IO_OUT)
+  else if (condition & G_IO_OUT)
     gnutls_source->stream = G_OBJECT (gnutls->priv->tls_ostream);
-  else
-    {
-      gnutls_source->base_direction = gnutls->priv->internal_direction;
-      gnutls_source->stream = NULL;
-    }
-  gnutls_source_sync_child_source (gnutls_source, gnutls_source->base_direction);
+  gnutls_source_sync_child_source (gnutls_source);
 
   if (cancellable)
     {
@@ -696,8 +801,10 @@ g_tls_connection_gnutls_pull_func (gnutls_transport_ptr_t  transport_data,
 
   if (ret < 0)
     set_gnutls_error (gnutls, G_IO_IN);
+#ifndef GNUTLS_E_PREMATURE_TERMINATION
   else if (ret == 0)
     gnutls->priv->eof = TRUE;
+#endif
 
   return ret;
 }
@@ -739,10 +846,13 @@ handshake_internal (GTlsConnectionGnutls  *gnutls,
 		    GCancellable          *cancellable,
 		    GError               **error)
 {
+  GTlsCertificate *peer_certificate = NULL;
+  GTlsCertificateFlags peer_certificate_errors = 0;
   int ret;
 
   if (G_IS_TLS_SERVER_CONNECTION_GNUTLS (gnutls) &&
-      gnutls->priv->ever_handshaked && !gnutls->priv->need_handshake)
+      gnutls->priv->ever_handshaked && !gnutls->priv->handshaking &&
+      !gnutls->priv->need_handshake)
     {
       BEGIN_GNUTLS_IO (gnutls, blocking, cancellable);
       ret = gnutls_rehandshake (gnutls->priv->session);
@@ -752,8 +862,23 @@ handshake_internal (GTlsConnectionGnutls  *gnutls,
 	return FALSE;
     }
 
-  gnutls->priv->handshaking = TRUE;
-  G_TLS_CONNECTION_GNUTLS_GET_CLASS (gnutls)->begin_handshake (gnutls);
+  if (!gnutls->priv->handshaking)
+    {
+      gnutls->priv->handshaking = TRUE;
+
+      if (gnutls->priv->peer_certificate)
+	{
+	  g_object_unref (gnutls->priv->peer_certificate);
+	  gnutls->priv->peer_certificate = NULL;
+	  gnutls->priv->peer_certificate_errors = 0;
+
+	  g_object_notify (G_OBJECT (gnutls), "peer-certificate");
+	  g_object_notify (G_OBJECT (gnutls), "peer-certificate-errors");
+	}
+
+      g_tls_connection_gnutls_set_handshake_priority (gnutls);
+      G_TLS_CONNECTION_GNUTLS_GET_CLASS (gnutls)->begin_handshake (gnutls);
+    }
 
   BEGIN_GNUTLS_IO (gnutls, blocking, cancellable);
   ret = gnutls_handshake (gnutls->priv->session);
@@ -781,16 +906,40 @@ handshake_internal (GTlsConnectionGnutls  *gnutls,
 	  for (i = num_certs - 1; i >= 0; i--)
 	    {
 	      cert = g_tls_certificate_gnutls_new (&certs[i], chain);
+	      if (chain)
+		g_object_unref (chain);
 	      chain = cert;
 	    }
 	}
 
-      g_tls_connection_set_peer_certificate (G_TLS_CONNECTION (gnutls), chain);
+      peer_certificate = chain;
+    }
+
+  if (peer_certificate)
+    {
+      if (!G_TLS_CONNECTION_GNUTLS_GET_CLASS (gnutls)->verify_peer (gnutls, peer_certificate, &peer_certificate_errors))
+	{
+	  g_object_unref (peer_certificate);
+	  g_set_error_literal (error, G_TLS_ERROR, G_TLS_ERROR_BAD_CERTIFICATE,
+			       _("Unacceptable TLS certificate"));
+	  return FALSE;
+	}
+    }
+
+  G_TLS_CONNECTION_GNUTLS_GET_CLASS (gnutls)->finish_handshake (gnutls, ret == 0, error);
+
+  if (ret == 0)
+    {
+      gnutls->priv->peer_certificate = peer_certificate;
+      gnutls->priv->peer_certificate_errors = peer_certificate_errors;
+
+      g_object_notify (G_OBJECT (gnutls), "peer-certificate");
+      g_object_notify (G_OBJECT (gnutls), "peer-certificate-errors");
+
+      return TRUE;
     }
   else
-    g_tls_connection_set_peer_certificate (G_TLS_CONNECTION (gnutls), NULL);
-
-  return G_TLS_CONNECTION_GNUTLS_GET_CLASS (gnutls)->finish_handshake (gnutls, ret == 0, error);
+    return FALSE;
 }
 
 static gboolean
@@ -876,6 +1025,8 @@ g_tls_connection_gnutls_handshake_async (GTlsConnection       *conn,
       g_simple_async_result_complete_in_idle (simple);
       g_object_unref (simple);
     }
+  else if (error)
+    g_error_free (error);
 
   source = g_tls_connection_gnutls_create_source (gnutls, 0, cancellable);
   g_source_set_callback (source,
@@ -979,9 +1130,6 @@ close_internal (GTlsConnectionGnutls  *gnutls,
 {
   int ret;
 
-  if (!gnutls->priv->require_close_notify)
-    return TRUE;
-
   /* If we haven't finished the initial handshake yet, there's no
    * reason to finish it just so we can close.
    */
